// VinciaMatching.h is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2015 P. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// This file contains header information for VINCIA's Matching class.
  
#ifndef VINCIAMATCHING_H
#define VINCIAMATCHING_H

#include "Pythia8/Basics.h"                   // Vec4
#include "Pythia8/Event.h"                    // Event
#include "VinciaAntenna.h"            // Antenna
#include "VinciaAntennaSet.h"         // AntennaSet
#include "VinciaCommon.h"             // Common
#include "VinciaResolution.h"         // Resolution 
#include "VinciaTools.h"              // VinClu
#include "VinciaMGInterface.h"        // MGInterface
#include "VinciaSubtractionTerms.h"   // Subtraction terms for NLO

using namespace Pythia8;

namespace Vincia {

  //***********************************************************************
  //
  // class Matching : for computing matching coefficients
  // 
  // Is structured like many other VINCIA classes, with a constructor that only
  // initializes a few init flags to false, so that we may know if the class
  // has not been properly initialized. There are three initialization routines 
  // which must be called before a Matching object can really be used: first 
  // initAntPtr() and initPtr() and then init(). The former two set the 
  // addresses to the various other objects that the matching class needs 
  // to function. In particular, the call to initAntPtr tells the matching class
  // where to look for the AntennaSet objects that will be used in the current 
  // run. The latter, init(), initializes the matching class itself, i.e., sets
  // its internal variables, etc. 
  // The class contains the following main usable functions 
  //    isMatched() : to tell whether there is any matching available for 
  //                  a given final state. 
  //    getME()     : to get a matrix element (squared)
  //    getPME()    : to give the matching coefficient for a given final
  //                  state, using the same shower functions as the rest
  //                  of the code (via the AntennaSet pointers). 
  //    polarize()  : to assign ME-based polarizations to a set of partons
  //    getP2to4()  : FOR DEVELOPMENT ONLY.
  //
  // More detailed descriptions are given below, with each function.
  //
  //***********************************************************************

  class Matching {
    
  public:
    
    // Constructor.
    Matching() {isInit=false; isInitPtr=false;};

    // Destructor.
    virtual ~Matching() {};

    // Initialize Pointers (must be done before init)
    void initPtr(MGInterface* mgInterfacePtrIn, 
		 ParticleData* particleDataPtrIn, Resolution* resolutionPtrIn, 
                 Rndm* rndmPtrIn, Settings* settingsPtrIn, 
		 SubtractionTerms* subtractionTermsPtrIn, VinClu* vinCluPtrIn, 
		 VinciaCommon* vinComPtrIn, Functions* functionsPtrIn) {
      mgInterfacePtr      = mgInterfacePtrIn;
      particleDataPtr     = particleDataPtrIn;
      resolutionPtr       = resolutionPtrIn;
      rndmPtr             = rndmPtrIn;
      settingsPtr         = settingsPtrIn; 
      subtractionTermsPtr = subtractionTermsPtrIn; 
      vinCluPtr           = vinCluPtrIn; 
      vinComPtr           = vinComPtrIn;
      functionsPtr        = functionsPtrIn;
      isInitPtr=true;
    }

    // Initialize Pointers to Antenna Sets (must be done before init)
    void initAntPtr(AntennaSet* antV0, AntennaSet* antV1=0, 
		    AntennaSet* antV2=0, AntennaSet* antV3=0) {
      antennaSets.resize(0);
      antennaSets.push_back(antV0);
      if (antV1 != 0) antennaSets.push_back(antV1);
      if (antV2 != 0) antennaSets.push_back(antV2);
      if (antV3 != 0) antennaSets.push_back(antV3);
    }
    void initAntSctPtr(AntennaSet* antV0, AntennaSet* antV1=0, 
		       AntennaSet* antV2=0, AntennaSet* antV3=0) {
      antennaSetsSct.resize(0);
      antennaSetsSct.push_back(antV0);
      if (antV1 != 0) antennaSetsSct.push_back(antV1);
      if (antV2 != 0) antennaSetsSct.push_back(antV2);
      if (antV3 != 0) antennaSetsSct.push_back(antV3);
    }
    
    // Initialize
    bool init(int nVariationsIn=1);

    // *********************************************************************
    // isMatched() : return true if any matching exists at this order
    // Default is most general case: unpolarized massive particles.
    // *********************************************************************

    bool isMatched(int nIn, vector<int> idParticles, bool isPolarized = false,
		   bool treatAsMassless = false);

    // *********************************************************************
    // isMatchedNLO() : return true if NLO matching exists at this order
    //        Note: polarization and mass effects are not yet implemented 
    //        for NLO matching, so here assumed unpolarized and massless
    // *********************************************************************

    bool isMatchedNLO(int nIn, vector<int> idParticles);

    // *********************************************************************
    // getME() : Return a tree-level Matrix Element (Squared)
    // *********************************************************************

    double getME(int nIn, vector<int> id, vector<Vec4> p, vector<double> pol);

    // Wrapper to use list of particles as input
    double getME(int nIn, vector<Particle> particles); 

    // *********************************************************************
    // getPME() : Return a matching coefficient
    // INPUT: 
    //      nIn : 1 (for decay) or 2 (for 2->2 process)
    //      particlePtrs : list of pointers to particles, incoming ones first
    //   Note   : the particle list is assumed to be color-ordered
    // OUTPUT: 
    //      PME : vector whose [0] component is the matching coefficient
    //            for the current shower settings, with the additional ones
    //            corresponding to uncertainty variations (if switched on).
    // *********************************************************************

    bool getPME(int nIn, vector<Particle> particles, vector<double>& PME);

    // *********************************************************************
    // polarize() : use helicity matrix elements to polarize a list of 
    // partons. Return true if successful, false otherwise. 
    // Note: is able to polarize both massless and massive parton states
    // *********************************************************************

    bool polarize(int nIn, vector<Particle>& particlePtrs); 
    
    // Wrapper to operate directly on particles in an event record
    bool polarize(Event& event, int nIn, vector<int> iParticle);

    // *********************************************************************
    // getVME() : Return the virtual correction to the matching coefficient
    // *********************************************************************

    // Get a vector of  V (one-loop) coefficients, for:
    //    mu2PS        = shower renormalization scale squared
    //    evVarEmit    = gluon evolution variable type (default = 2pT)
    //    smoothPower  = 0 (strong ordering) 0.5 ("stronger") 1.0 (smooth) 
    //    splitUsesMqq = true if g->qq uses mqq as evolution variable 
    //    splitUsesAri = sets whether Ariadne factor is applied to g->qq
    bool getVME(int nIn, vector<Particle> particles, 
                    vector<double>& V1);
                 
    // *********************************************************************
    // getP2to4() : Return a 2->4 matching coefficient (under development)
    // *********************************************************************

    bool getP2to4(vector<int> id, vector<Vec4> p, vector<double>& PME);

    // *********************************************************************
    // getAntTree() : return |M_n|^2/|M_Born|^2 for these daughters
    // Not used for the time being so commented out
    // *********************************************************************

    // double getAntTree(int idRes, vector<Particle> daughters);

    // Wrapper function to return a specific antenna
    // inside a specific antenna set
    // 0: User, 1: Def, 2: Max, 3: Min    
    Antenna* getAnt(int iAnt, bool isSectorIn=false, int iSet=0) {
      if (!isSectorIn) 
        return antennaSets[iSet]->getAnt(iAnt);
      else 
        return antennaSetsSct[iSet]->getAnt(iAnt);
    }
    
    // Print internal and diagnostic histrograms
    void printHistos( ostream& os = cout ) {
      map<string,Hist>::iterator iH;
      for (iH=HrDeltaA.begin();iH!=HrDeltaA.end();++iH) {	
	string Hname=iH->first;
        if ( HrDeltaA[Hname].getEntries() >= 1) {
	  os<<Hname<<HrDeltaA[Hname]<<endl;
	}
      }
    }
    
    // Print internal and diagnostic histrograms
    void writeHistos( string fileName="vincia", string lastName="dat" ) {
      // Now print histograms in shower class
      map<string,Hist>::iterator iH;
      int hNum = 0;
      for (iH=HrDeltaA.begin();iH!=HrDeltaA.end();++iH) {	
	string Hname=iH->first;
        if ( HrDeltaA[Hname].getEntries() >= 1) {
          ++hNum;
          string file = fileName+"-dA"+num2str(hNum,int(log10(hNum)+1))
            +"."+lastName;
          HrDeltaA[Hname].table(file);
	}
      }
    }

    // Get info from last call
    int getLastNordered() {return nOrdered;}
    vector<double> getLastQ2() {return vecQ2;}
    vector<double> getLastQ2hat() {return vecQ2hat;}
    
    // Set verbose level
    void setVerbose(int verboseIn) {verbose = verboseIn;}

    // Set ordering Mode
    void setOrderingMode(int orderingModeIn) {orderingMode = orderingModeIn;}
    
  private:

    // Is initialized?    
    bool isInit, isInitPtr;

    // Is MG or BH Matching switched on?
    bool isOnMG;
    bool isOnBH;

    // Uncertainty Bands
    bool uncertaintyBands;
    int nVariations;

    // Pointers to PYTHIA and VINCIA objects
    Resolution*        resolutionPtr;
    MGInterface*       mgInterfacePtr;
    ParticleData*      particleDataPtr;
    Rndm*              rndmPtr;
    Settings*          settingsPtr;
    VinClu*            vinCluPtr;
    VinciaCommon*      vinComPtr;
    Functions*         functionsPtr;

    // Vector of pointers to antenna sets (0 is user, >= 1 are variations)
    vector<AntennaSet*> antennaSets;    
    vector<AntennaSet*> antennaSetsSct; 

    // Subtraction terms: owned by VinciaMatching
    SubtractionTerms* subtractionTermsPtr;

    // Evolution Mode
    int evolutionType, orderingMode, evolutionTypeDef, orderingModeDef;
    double pTnorm, mDnorm, pTnormDef, mDnormDef;

    // Numerical settings
    int nPointsMCmin, nPointsMCmax;

    // AlphaS
    AlphaStrong* alphaSptr;
    double alphaSmuFreeze, alphaSmuFreezeDef;
    double alphaSkMu, alphaSkMuDef, alphaSmax;
    int alphaSmode, alphaSmodeDef, alphaSmodeNLO;
    bool alphaScmw, alphaScmwDef;

    // Sector ordering
    bool sectorShower;

    // Verbose level
    int verbose;

    // Saved info about last getPME call that can be used for later study
    int nOrdered;
    vector<double> vecQ2;
    vector<double> vecQ2hat;
    
    // Histograms for runtime statistics
    map<string,Hist> HrDeltaA;
  };

  // Class for cached matrix elements 
  // (used by unpolarized sector shower to speed up matching)

  class MEcache {

  public:
    
    //Constructor
    MEcache() {};
    
    // 
    bool hasCache(int n) {
      if (n <= int(vCacheSav.size())) return hasCacheSav[n];
      else return false;
    }

    //
    void setCache(int n, double val) {
      // Fill up any intermediate blanks with zeroes and false
      for (int i=vCacheSav.size(); i<=n; ++i) {
	vCacheSav.push_back(0.0);
	hasCacheSav.push_back(false);
      }
      // Cache value
      vCacheSav[n] = val;
      hasCacheSav[n] = true;
    }

    // 
    double getCache(int n) {
      if (hasCache(n)) return vCacheSav[n];
      else return -1.0;
    }

    // 
    void resetCache(int iCache=-1) {
      // Default: reset all elements
      if (iCache <= -1) {
	vCacheSav.resize(0);
	hasCacheSav.resize(0);
      } 
      // i > 0 : reset a specific element
      else if (iCache <= int(vCacheSav.size()) - 1) {
	vCacheSav[iCache]   = 0.0;
	hasCacheSav[iCache] = false;
      }
      // if i > size of cache, add empty/false elements up to i
      else {
	for (int i=vCacheSav.size(); i<=iCache; ++i) {
	  vCacheSav.push_back(0.0);
	  hasCacheSav.push_back(false);
	}	
      }
    }

  private:

    vector<int> vCacheSav;
    vector<bool> hasCacheSav;    

  };

}

#endif

